/* bzflag
 * Copyright (c) 1993 - 2005 Tim Riker
 *
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the license found in the file
 * named COPYING that should have accompanied this file.
 *
 * THIS PACKAGE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

// bzfscron plugin - cron-style timed command execution for bzfs
// see ReadMe.txt for usage notes.

#include "common.h"

#include <string>
#include <vector>
#include <fstream>
#include <math.h>

#include "CronJob.h"
#include "bzfsAPI.h"
#include "TextUtils.h"

// FIXME: these should go away.
// depending on network communication with the server is stupid
// considering that we're, like, the same process as it is
// but until we have some sort of "fake player" support, it will have to do
#include "Address.h"
#include "SynchroServerLink.h"
int debugLevel;

// -- Event handler class

class CronManager : public bz_EventHandler {
public:
  CronManager();
  ~CronManager();

  void process(bz_EventData *eventData);
  bool autoDelete(void) {return false;};

  bool connect();
  bool connect2();
  bool connect3();
  void sendMessage(std::string message);

  void addJob(CronJob &job) {jobs.push_back(job);};
  void clearJobs() {jobs.clear();};

private:
  std::vector<CronJob> jobs;
  double lastTick;
  int lastMinute;

  double startupTime;
  double stageTime;
  int stage;

  int cronID;
  void getNet();
  ServerLink* sLink;
};

CronManager cron;

// -- Library entry points for bzfs

#include "bzfsAPI.h"

BZ_GET_PLUGIN_VERSION

BZF_PLUGIN_CALL int bz_Load(const char* commandLine)
{
  // should have a filename on the command line.  try to open it.
  if (!commandLine) {
    bz_debugMessage(1, "bzfscron: no crontab specified");
    return 1;
  }
  std::ifstream input(commandLine, std::ios::in);
  if (input.peek() == EOF) {
    bz_debugMessage(1, "bzfscron: crontab nonexistant or invalid");
    return 1;
  }

  // set the debug level
  debugLevel = bz_getDebugLevel();

  // read in the crontab
  char buffer[1024];
  while (!input.eof() && !input.fail() && input.good()) {
    input.getline(buffer, sizeof(buffer));
    if (buffer[0] != '#') {
      CronJob newcron(buffer);
      cron.addJob(newcron);
    }
  }

  // we have a granularity of 1 minute but we want to run things ASAP in that minute
  // since we rely on "real time" and not "relative time". n+<=5 sec should not hurt.
  bz_setMaxWaitTime(5.0f);

  // but network relies on delays of <=100ms (FIXME)
  bz_setMaxWaitTime(0.1f);

  // register to receive ticks
  bz_registerEvent(bz_eTickEvent, &cron);

  bz_debugMessage(4, "bzfscron 0.2: plugin loaded");
  return 0;
}

BZF_PLUGIN_CALL int bz_Unload(void)
{
  bz_removeEvent(bz_eTickEvent, &cron);
  bz_debugMessage(4, "bzfscron 0.2: plugin unloaded");
  return 0;
}


// -- Utility functions

// calculate the day of the week (0 = sunday)
static int dow(int m, int d, int y) {
  // W = (k + floor(2.6m - 0.2) - 2C + Y + floor(Y/4) + floor(C/4)) mod 7

  m -= 2; // march is base month
  if (m < 1) {
    y -= 1;
    m += 12;
  }

  const int c = (int)((y - 50.0f) / 100.0f); // century

  // whew!
  return ((int)(d + floor(2.6f * m - 0.2f) - 2 * c + y + floor(y / 4.0f) + floor(c / 4.0f)) % 7);
}


// -- Event callbacks

CronManager::CronManager() : lastTick(0.0), lastMinute(-1), stage(0) {
  // do nothing
}

CronManager::~CronManager() {
  delete sLink;
#ifdef WIN32
  WSACleanup();
#endif
}

void CronManager::process(bz_EventData *eventData) {
  switch (eventData->eventType) {
  
  default: {
    // no, sir, we didn't ask for THIS!!
    bz_debugMessage(1, "bzfscron: received event with eventType != bz_eTickEvent!");
    return;
  }

  case bz_eTickEvent: {
    bz_TickEventData* event = (bz_TickEventData*)eventData;

    // initialize the first tick through
    if (lastTick == 0.0f) {
      lastTick = startupTime = stageTime = event->time;
      bz_debugMessage(3, "bzfscron: initialized timer system");
      return;
    }

    // connection stuff 30 seconds after startup, or 5 seconds after the last try.
    // server should be listening by now.
    if (stage == 0) {
      if (event->time > startupTime + 30.0f && event->time > stageTime + 5.0f) {
	stageTime = event->time;
	bz_debugMessage(4, "bzfscron: connecting, stage 1.");
	// activate myself.  stage 1 connection, establish socket link.
	if (!cron.connect()) {
	  bz_debugMessage(1, "bzfscron: stage 1 connection to server failed");
	  return;
	} else {
	  stage = 1;
	}
      }
      return;
    } else if (stage == 1) {
      // wait a couple ticks for synchronicity's sake.
      // remember we're the SAME THREAD as bzfs is, this gets real ugly real fast
      // we need to wait a tick between EVERY tx and rx otherwise bzfs never has a
      // chance to process what we send, and we end up blocking for eternity
      if (event->time > stageTime + 0.5f) {
	stageTime = event->time;
	bz_debugMessage(4, "bzfscron: connecting, stage 2.");
	// stage 2 connection.  check version, not banned, etc.
	if (!cron.connect2()) {
	  bz_debugMessage(1, "bzfscron: stage 2 connection to server failed");
	  stage = 0;
	  return;
	} else {
	  stage = 2;
	}
      }
      return;
    } else if (stage == 2) {
      if (event->time > stageTime + 0.5f) {
	stageTime = event->time;
	bz_debugMessage(4, "bzfscron: connecting, stage 3.");
	// stage 3 connection.  send MsgEnter, finalize.
	if (!cron.connect3()) {
	  bz_debugMessage(1, "bzfscron: stage 3 connection to server failed");
	  stage = 0;
	  return;
	} else {
	  stage = 3;
	}
      }
      return;
    }

    // perform network junk
    getNet();

    // ignore ticks that are less than 5 seconds apart
    if (lastTick + 4.95f > event->time)
      return;
    lastTick = event->time;
    bz_debugMessage(4, "bzfscron: tick!");

    // ensure that the minute has changed
    bz_localTime t;
    bz_getLocaltime(&t);
    if (t.minute == lastMinute)
      return;
    lastMinute = t.minute;
    bz_debugMessage(4, "bzfscron: minute change");

    // iterate through all the jobs.  if they match the current minute, run them.
    std::vector<CronJob>::iterator itr;
    for (itr = jobs.begin(); itr != jobs.end(); ++itr)
      if (itr->matches(t.minute, t.hour, t.day, t.month, dow(t.month, t.day, t.year))) {
	bz_debugMessage(4, "bzfscron: job matched");
	sendMessage(itr->getCommand());
      }

    break;
  }
  }
}


// WHOA!!!! This is VERY VERY UGLY.  Don't read this part.  No really, don't do it.
// FIXME: bzfs sometime in the future should allow the addition of a "fake player" that would
// make all this junk unnecessary

bool CronManager::connect() {
#ifdef _WIN32
  // startup winsock
  {
    static const int major = 2, minor = 2;
    WSADATA wsaData;
    if (WSAStartup(MAKEWORD(major, minor), &wsaData)) {
      std::cerr << "Could not initialise WinSock.";
      return false;
    }
    if (LOBYTE(wsaData.wVersion) != major ||
	HIBYTE(wsaData.wVersion) != minor) {
      std::cerr << "Invalid WinSock version (got " << (int) LOBYTE(wsaData.wVersion) <<
	'.' << (int) HIBYTE(wsaData.wVersion) << ", expected" << major << '.' << minor << ')';
      WSACleanup();
      return false;
    }
  }
#endif

  std::string server = bz_getPublicAddr().c_str();
  int port = 5154;
  int n = (int)server.rfind(":", server.size());
  if (n != std::string::npos && n != server.size()) port = atoi(server.substr(n + 1).c_str());
  bz_debugMessage(4, TextUtils::format("bzfscron: connecting to %s, port %d", "127.0.0.1", port).c_str());

  sLink = new ServerLink(Address("127.0.0.1"), port);

  if (sLink->getState() != ServerLink::Connecting) {
    bz_debugMessage(1, "bzfscron: unable to establish connection.");
    return false;
  }

  return true;
}

bool CronManager::connect2() {
  sLink->stage2();

  cronID = sLink->getId();
  bz_debugMessage(4, TextUtils::format("bzfscron: got player ID %d", cronID).c_str());

  if (sLink->getState() != ServerLink::Okay) {
    bz_debugMessage(1, "bzfscron: rejected by server.");
    return false;
  }

  sLink->sendEnter(TankPlayer, ObserverTeam, "bzfscron", "", "");

  return true;
}

bool CronManager::connect3() {
  if (sLink->getState() != ServerLink::Okay) {
    bz_debugMessage(1, "bzfscron: rejected by server.");
    return false;
  }

  std::string reason;
  uint16_t code, rejcode;
  if (!sLink->readEnter(reason, code, rejcode)) {
    bz_debugMessage(1, TextUtils::format("bzfscron: rejected by server (with code %d).", rejcode).c_str());
    return false;
  }

  // I think I'll make myself admin, so I can run all sorts of fun commands
  if (!bz_setPlayerAdmin(cronID)) {
    bz_debugMessage(1, "bzfscron: unable to make myself an administrator");
    return false;
  }

  // But nobody needs to know I'm admin, 'cause I can't do anything unless crontab told me to
  bz_grantPerm(cronID, bz_perm_hideAdmin);

  return true;
}

void CronManager::getNet() {
  uint16_t code, len;
  char inbuf[MaxPacketLen];

  // read the data so we don't end up with network errors
  if (sLink->read(code, len, inbuf) == 1) {
    // we don't handle very much
    switch (code) {
      case MsgSuperKill:
	// FIXME: we should handle this case by reconnecting next tick
	break;
      default:
	break;
    }
  }
}

void CronManager::sendMessage(std::string msg) {
  char buffer[MessageLen];
  char buffer2[1 + MessageLen];
  void* buf = buffer2;

  buf = nboPackUByte(buf, AllPlayers);
  memset(buffer, 0, MessageLen);
  strncpy(buffer, msg.c_str(), MessageLen - 1);
  nboPackString(buffer2 + 1, buffer, MessageLen);
  sLink->send(MsgMessage, sizeof(buffer2), buffer2);
}


// Local Variables: ***
// mode:C++ ***
// tab-width: 8 ***
// c-basic-offset: 2 ***
// indent-tabs-mode: t ***
// End: ***
// ex: shiftwidth=2 tabstop=8

